/****************************************************************************
Copyright (c) 2012,Otto-von-Guericke University of Magdeburg. All rights 
reserved.

This program and accompanying materials are made available under the terms of
the Eclipse Public License - v 1.0, http://www.eclipse.org/legal/epl-v10.html
****************************************************************************/
package index.structures;

import index.base.AIndexStructure;

import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.HashMap;

import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JPanel;

import metrics.DistanceMetric;

import testapp.Dictionary;
import testapp.Dictionary.Bundle;
import testapp.JOptionDialog;
import testapp.tapedpane.JTappedPane_Panel;


public class IndexVA_SSA extends AIndexStructure {
	private int bitsPerDimension = 4;

	private int numberOfDimensions;
	private int[][] approximationFile_Int;
	private int count;
	private int rangeOfCell;
	
	private DistanceMetric metric = DistanceMetric.getInstance();

	public IndexVA_SSA(int _size, int _dimension, int _minValue, int _maxValue,
			HashMap<String, Object> map) {
		super(_size, _dimension, _minValue, _maxValue, map);
		initParameter(map);

		numberOfDimensions = _dimension;

		approximationFile_Int = new int[_size][intsPerApproximation()];

		count = 0;

		rangeOfCell = (_maxValue - _minValue + 1) / power(2, bitsPerDimension);
		
	}

	private int intsPerApproximation() {
		return (numberOfDimensions * bitsPerDimension + 31) / 32;
	}

	private int power(int base, int exp) {
		if (exp == 1) {
			return base;
		}
		int tmp = power(base, exp / 2);
		if (exp % 2 == 0) {
			return tmp * tmp;
		} else {
			return tmp * tmp * base;
		}
	}

	@Override
	public boolean canKNNSearch() {
		return true;
	}

	@Override
	public void clear() {
		approximationFile_Int = null;
		count = 0;
	}

	@Override
	public String getInformations() {
		return getName()+" \n" + Dictionary.get(Bundle.INDEX,"bitsperdimension")
				+ bitsPerDimension;
	}

	@Override
	public boolean insert(int[] dataset) {

		int[] coordinate_Int = encodeApprox(approximation((dataset)));
		approximationFile_Int[count] = coordinate_Int;
		
		count++;
		return true;
	}

	@Override
	public int[] search(int[] datasetToSearch) {
		int[] approx = encodeApprox(approximation(datasetToSearch));
		int IOops = 0;
		for (int c0 = 0; c0 < count; c0++) {
			if (Arrays.equals(approximationFile_Int[c0], approx)) {
				IOops++;
				// einmal getDataFromDisk
				int[] curPoint = getDataset(c0);
				
				boolean isIdentity = true;
	            for (int k = 0; k < curPoint.length; k++) {
	                if (curPoint[k] != datasetToSearch[k]) {
	                    isIdentity = false;
	                    break;
	                }
	            }
				if(isIdentity)
					return curPoint;
				
				
//				if (Arrays.equals(curPoint, datasetToSearch)) {
//					return curPoint;
//				}
			}
		}
		return null;
	}

	public int[][] searchKNN(int[] datasetToSearch, int k) {
		double delta = Double.MAX_VALUE;

		int[][] ans = new int[k][dim];
		double[] dst = new double[k];
		for (int c0 = 0; c0 < dst.length; c0++) {
			dst[c0] = Double.MAX_VALUE;
		}


		for (int c0 = 0; c0 < count; c0++) {
			double minDist = getLowerBound(approximation(datasetToSearch),
					datasetToSearch, decodeApprox(approximationFile_Int[c0]));
			
			if (minDist < delta) {
				int[] cand = getDataset(c0);
				double d = metric.calculateDistance(datasetToSearch, cand);
				if (d < delta) {
					for (int c1 = 0; c1 < k; c1++) {
						if (d < dst[c1]) {
							for (int c2 = k - 2; c2 >= c1; c2--) {
								dst[c2 + 1] = dst[c2];
								ans[c2 + 1] = ans[c2];
							}
							dst[c1] = d;
							ans[c1] = cand;
							delta = dst[k - 1];
							break;
						}
					}
				}
			}
		}

		return ans;
	}

	private int[] encodeApprox(int[] in) {
		int[] approx = new int[intsPerApproximation()];
		int shortIndex = 0;
		int bitsAvailable = 32;
		for (int i = 0; i < in.length; i++) {
			int bitsToStore = bitsPerDimension;
			int bits = in[i];
			while (bitsToStore > 0) {
				int storeBits = Math.min(bitsToStore, bitsAvailable);
				approx[shortIndex] <<= storeBits;
				int tBits = bits >> (bitsPerDimension - storeBits);
				approx[shortIndex] |= tBits;
				bits -= tBits;

				bitsToStore -= storeBits;
				bitsAvailable -= storeBits;
				if (bitsAvailable == 0) {
					shortIndex++;
					bitsAvailable = 32;
				}
			}
		}
		approx[approx.length - 1] <<= bitsAvailable;
		return approx;
	}

	private int[] approximation(int[] vector) {
		int[] approx = new int[vector.length];
		for (int c0 = 0; c0 < vector.length; c0++) {
			approx[c0] = vector[c0] / rangeOfCell;
		}
		return approx;
	}

	private int[] decodeApprox(int[] approx) {
		int[] res = new int[dim];

		int approxPointer = approx.length - 1;
		int stuffedBits = 32 - ((dim * bitsPerDimension) % 32);

		int currentApprox = approx[approxPointer];
		
		int bitsInApproxLeft = 32;

		if(	stuffedBits != 32){
			approxPointer--;
			currentApprox >>= stuffedBits;
			bitsInApproxLeft = 32 - stuffedBits;	
		}
		
		for (int c0 = res.length - 1; c0 >= 0; c0--) {
			int bitsToGetLeft = bitsPerDimension;
			while (bitsToGetLeft > 0) {
				if (bitsInApproxLeft == 0) {
					currentApprox = approx[approxPointer];
					approxPointer--;
					bitsInApproxLeft = 32;
				}
				
				int getBits = Math.min(bitsToGetLeft, bitsInApproxLeft);
				res[c0] = (currentApprox << (32 - getBits)) >>> (32 - getBits);
				res[c0] <<= bitsToGetLeft - getBits;
				currentApprox >>>= getBits;
				bitsToGetLeft -= getBits;
				bitsInApproxLeft -= getBits;
			}
		}
		return res;
	}

	private double getLowerBound(int[] qApprox, int[] qData, int[] pApprox) {
//		double dist = 0;
//		for (int c0 = 0; c0 < qApprox.length; c0++) {
//			if (qApprox[c0] < pApprox[c0]) {
//				dist += Math.pow(pApprox[c0] * rangeOfCell - qData[c0], 2);
//			} else if (qApprox[c0] > pApprox[c0]) {
//				dist += Math.pow(qData[c0] - ((pApprox[c0] + 1) * rangeOfCell),
//						2);
//			}
//		}
		int[] t = new int[pApprox.length];
		int[] t2 = new int[pApprox.length]; 
		    Array.setInt(t2, 0, 0);
		for (int c0 = 0; c0 < pApprox.length; c0++) {
			
			if (qApprox[c0] < pApprox[c0]) {
				t[c0] = pApprox[c0] * rangeOfCell - qData[c0];
			} else if (qApprox[c0] > pApprox[c0]) {
				t[c0] = qData[c0] - ((pApprox[c0] + 1) * rangeOfCell);
			}	
		}
		return metric.calculateDistance(t, t2);
	}

	public JOptionDialog getOptionsDialog(JFrame parent) {
		return new VA_SSA_OptionDialog(parent, this);
	}

	private class VA_SSA_OptionDialog extends JOptionDialog {
		private static final long serialVersionUID = 1L;

		JFormattedTextField field;
		JTappedPane_Panel northPanel;

		public VA_SSA_OptionDialog(JFrame parent, AIndexStructure is) {
			super(parent, is.getName());
		}

		public JPanel getNorthPanel() {
			northPanel = new JTappedPane_Panel(null) {

				private static final long serialVersionUID = 1L;

				@Override
				public void refreshResource() {

				}

			};

			field = northPanel.getFormtedTextFieldInstance_Integer(3);
			field.setValue(bitsPerDimension);
			field.setColumns(2);
			field.setToolTipText("4-6");

			northPanel.addSimpleLabelAndInputRow(0, Dictionary.get(Bundle.INDEX,"bitsperdimension")+":",
					field);

			return northPanel;
		}

		@Override
		public HashMap<String, Object> getOptions() {
			HashMap<String, Object> map = new HashMap<String, Object>();
			map.put("bitsPerDimension",
					"" + field.getValue());
			return map;
		}
	}

	public String getName() {
		return Dictionary.get(Bundle.INDEX,"vassa");
	}
}
